# -*- ruby -*-
# frozen_string_literal: true

require 'cztop/reactor' unless defined?( CZTop::Reactor )

# Methods for logging monitor events.
module CZTop::Reactor::SocketMonitoring

	### Inclusion callback -- add Loggability to including classes.
	def self::included( mod )
		super
		mod.extend( Loggability )
		mod.log_to( :cztop ) unless Loggability.log_host?( mod )
	end


	### Set up a monitor instance variable on object creation.
	def initialize( * ) # :notnew:
		super if defined?( super )
		@monitor = nil
	end

	##
	# The CZTop::Monitor for the server socket
	attr_reader :monitor


	### Run the block with monitoring for the specified +socket+.
	def with_socket_monitor( reactor, socket, *events )
		mon = self.set_up_socket_monitor( reactor, socket, *events )

		return yield
	ensure
		self.clean_up_socket_monitor( mon )
	end


	### Create a monitor for the specified +socket+.
	def set_up_socket_monitor( reactor, socket, *events )
		return reactor.register_monitor( socket, *events, &self.method(:on_monitor_event) )
	end


	### Tear down the monitor.
	def clean_up_socket_monitor( mon )
		mon.terminate if mon
		self.log.debug "Monitor is: %p" % [ mon ]
	end


	### Handle monitor events.
	def on_monitor_event( monitor_event )
		# self.log.debug "Got monitor event: %p" % [ monitor_event ]

		msg = monitor_event.socket.receive
		type, *payload = *msg
		callback_name = "on_#{type.downcase}"

		if self.respond_to?( callback_name, true )
			self.send( callback_name, *payload )
		else
			self.log.warn "No handler (#%s) for monitored %s event." % [ callback_name, type ]
		end
	end
	alias_method :handle_monitor_event, :on_monitor_event


	#########
	protected
	#########


	### Monitor event callback for socket connection events
	def on_connected( fd, endpoint )
		self.log.debug "Client socket on FD %d connected" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for socket connection-delayed events
	def on_connect_delayed( fd, endpoint )
		self.log.debug "Client socket on FD %d connection delayed" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for socket retry events
	def on_connect_retried( fd, endpoint )
		self.log.debug "Retrying connection for socket on FD %d" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for listening events
	def on_listening( fd, endpoint )
		self.log.debug "Socket on FD %d is now listening for connections" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for bind failure events
	def on_bind_failed( fd, endpoint )
		self.log.debug "Socket on FD %d could not be bound" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for socket accepted events
	def on_accepted( fd, endpoint )
		self.log.debug "Client socket on FD %d accepted" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for socket accept failed events
	def on_accept_failed( fd, endpoint )
		self.log.debug "Client socket on FD %d accept failed" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for socket closed events
	def on_closed( fd, endpoint )
		self.log.debug "Client socket on FD %d closed" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for socket close failed events
	def on_close_failed( fd, endpoint )
		self.log.debug "Close failed for socket on FD %d" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for socket disconnection events
	def on_disconnected( fd, endpoint )
		self.log.debug "Client socket on FD %d disconnected" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for monitor stopped events.
	def on_monitor_stopped( fd, endpoint )
		self.log.debug "Monitor for socket on FD %d stopped." % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for successful auth events.
	def on_handshake_succeeded( fd, endpoint )
		self.log.debug "Client socket on FD %d handshake succeeded" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for failed handshake events.
	def on_handshake_failed_no_detail( fd, endpoint )
		self.log.debug "Client socket on FD %d handshake failed; no further details are known" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for failed handshake events.
	def on_handshake_failed_protocol( fd, endpoint )
		self.log.debug "Client socket on FD %d handshake failed: protocol error" % [ fd ]
		super if defined?( super )
	end


	### Monitor event callback for failed auth events.
	def on_handshake_failed_auth( fd, endpoint )
		self.log.debug "Client socket on FD %d handshake failed: authentication failure" % [ fd ]
		super if defined?( super )
	end

end # module CZTop::Reactor::SocketMonitoring
